from Accuinsight.modeler.entities._modeler_object import _ModelerObject


class LcRunInfo(_ModelerObject):
    """
    Metadata about a run.
    """

    def __init__(self,
                 user_id,
                 start_time,
                 end_time,
                 duration,
                 artifact_location,
                 run_status,
                 path,
                 model_path,
                 json_path,
                 run_name=None,
                 note=None):
        if user_id is None:
            raise Exception("user_id cannot be None")
        if start_time is None:
            raise Exception("start_time cannot be None")
        if end_time is None:
            raise Exception("end_time cannot be None")
        if duration is None:
            raise Exception("duration cannot be None")
        if artifact_location is None:
            raise Exception("duratiartifact_locationn cannot be None")
        if run_status is None:
            raise Exception("run_status cannot be None")
        if path is None:
            raise Exception("notebook_path cannot be None")
        if model_path is None:
            raise Exception("model file path cannot be None")
        if json_path is None:
            raise Exception("model json file path cannot be None")

        self._run_name = run_name
        self._user_id = user_id
        self._start_time = start_time
        self._end_time = end_time
        self._duration = duration
        self._artifact_location = artifact_location
        self._run_status = run_status
        self._note = note
        self._path = path
        self._model_path = model_path
        self._json_path = json_path

    def __eq__(self, other):
        if type(other) is type(self):
            # TODO deep equality here?
            return self.__dict__ == other.__dict__
        return False

    @property
    def run_name(self):
        """String the name of run."""
        return self._run_name

    @property
    def user_id(self):
        """String ID of the user who initiated this run."""
        return self._user_id

    @property
    def start_time(self):
        """Start time of the run, in number of milliseconds since the UNIX epoch."""
        return self._start_time

    @property
    def end_time(self):
        """End time of the run, in number of milliseconds since the UNIX epoch."""
        return self._end_time

    @property
    def duration(self):
        """The total elapsed time of the run, in number of milliseconds since the UNIX epoch."""
        return self._duration

    @property
    def artifact_location(self):
        """The artifact location for the current run."""
        return self._artifact_location

    @property
    def run_status(self):
        """
        One of the values in /entities/lc_run_status.py
        describing the status of the run.
        """
        return self._run_status

    @property
    def note(self):
        """
        The extra information for the current run.
        This value is set by user.
        """
        return self._note

    @property
    def path(self):
        """
        the name of current selected notebook.
        """
        return self._path

    @property
    def model_path(self):
        """
        the path of model file.
        """
        return self._model_path

    @property
    def json_path(self):
        """
        the path of model json file.
        """
        return self._json_path

    def to_proto(self):
        proto = ProtoLcRunInfo()
        proto.name = self.run_name
        proto.userId = self.user_id
        proto.creaDt = self.start_time
        proto.endDt = self.end_time
        proto.duration = self.duration
        proto.afLoc = self.artifact_location
        proto.status = RunStatus.from_string(self.run_status)
        proto.note = self.note
        proto.path = self._path
        proto.modelPath = self._model_path
        proto.jsonPath = self._json_path
        return proto

    @classmethod
    def from_proto(cls, proto):
        end_time = proto.endDt
        # The proto2 default scalar value of zero indicates that the run's end time is absent.
        # An absent end time is represented with a NoneType in the `RunInfo` class
        if end_time == 0:
            end_time = None
        return cls(user_id=proto.userId,
                   start_time=proto.creaDt,
                   end_time=proto.endDt,
                   duration=proto.duration,
                   artifact_location=proto.afLoc,
                   run_status=proto.status,
                   model_path=proto.modelPath,
                   json_path=proto.jsonPath,
                   run_name=proto.name,
                   note=proto.note)
from Accuinsight.modeler.entities.lc_run_status import RunStatus
from Accuinsight.modeler.entities._modeler_object import _ModelerObject
from Accuinsight.modeler.protos.life_cycle_pb2 import LcRunInfo as ProtoLcRunInfo


class LcRunInfo(_ModelerObject):
    """
    Metadata about a run.
    """

    def __init__(self,
                 user_id,
                 start_time,
                 end_time,
                 duration,
                 artifact_location,
                 run_status,
                 path,
                 model_path,
                 json_path,
                 run_name=None,
                 note=None):
        if user_id is None:
            raise Exception("user_id cannot be None")
        if start_time is None:
            raise Exception("start_time cannot be None")
        if end_time is None:
            raise Exception("end_time cannot be None")
        if duration is None:
            raise Exception("duration cannot be None")
        if artifact_location is None:
            raise Exception("duratiartifact_locationn cannot be None")
        if run_status is None:
            raise Exception("run_status cannot be None")
        if path is None:
            raise Exception("notebook_path cannot be None")
        if model_path is None:
            raise Exception("model file path cannot be None")
        if json_path is None:
            raise Exception("model json file path cannot be None")
        
        self._run_name = run_name
        self._user_id = user_id
        self._start_time = start_time
        self._end_time = end_time
        self._duration = duration
        self._artifact_location = artifact_location
        self._run_status = run_status
        self._note = note
        self._path = path
        self._model_path = model_path
        self._json_path = json_path

    def __eq__(self, other):
        if type(other) is type(self):
            # TODO deep equality here?
            return self.__dict__ == other.__dict__
        return False

    @property
    def run_name(self):
        """String the name of run."""
        return self._run_name

    @property
    def user_id(self):
        """String ID of the user who initiated this run."""
        return self._user_id

    @property
    def start_time(self):
        """Start time of the run, in number of milliseconds since the UNIX epoch."""
        return self._start_time

    @property
    def end_time(self):
        """End time of the run, in number of milliseconds since the UNIX epoch."""
        return self._end_time

    @property
    def duration(self):
        """The total elapsed time of the run, in number of milliseconds since the UNIX epoch."""
        return self._duration

    @property
    def artifact_location(self):
        """The artifact location for the current run."""
        return self._artifact_location

    @property
    def run_status(self):
        """
        One of the values in /entities/lc_run_status.py
        describing the status of the run.
        """
        return self._run_status

    @property
    def note(self):
        """
        The extra information for the current run.
        This value is set by user.
        """
        return self._note

    @property
    def path(self):
        """
        the name of current selected notebook.
        """
        return self._path

    @property
    def model_path(self):
        """
        the path of model file.
        """
        return self._model_path

    @property
    def json_path(self):
        """
        the path of model json file.
        """
        return self._json_path

    def to_proto(self):
        proto = ProtoLcRunInfo()
        proto.name = self.run_name
        proto.userId = self.user_id
        proto.creaDt = self.start_time
        proto.endDt = self.end_time
        proto.duration = self.duration
        proto.afLoc = self.artifact_location
        proto.status = RunStatus.from_string(self.run_status)
        proto.note = self.note
        proto.path = self._path
        proto.modelPath = self._model_path
        proto.jsonPath = self._json_path
        return proto

    @classmethod
    def from_proto(cls, proto):
        end_time = proto.endDt
        # The proto2 default scalar value of zero indicates that the run's end time is absent.
        # An absent end time is represented with a NoneType in the `RunInfo` class
        if end_time == 0:
            end_time = None
        return cls(user_id=proto.userId,
                   start_time=proto.creaDt,
                   end_time=proto.endDt,
                   duration=proto.duration,
                   artifact_location=proto.afLoc,
                   run_status=proto.status,
                   model_path=proto.modelPath,
                   json_path=proto.jsonPath,
                   run_name=proto.name,
                   note=proto.note)



#!/usr/bin/env python
# coding=utf-8

import functools
import hashlib
import os.path
import re
import sys
from pathlib import Path
from git import Repo, InvalidGitRepositoryError
import importlib
import pkg_resources
import json

from Accuinsight.modeler.utils.dependency.settings import SETTINGS
from Accuinsight.modeler.core.LcConst import LcConst


MB = 1048576
MODULE_BLACKLIST = set(sys.builtin_module_names)
# sadly many builtins are missing from the above, so we list them manually:
MODULE_BLACKLIST |= {
    None,
    "__future__",
    "_abcoll",
    "_bootlocale",
    "_bsddb",
    "_bz2",
    "_codecs_cn",
    "_codecs_hk",
    "_codecs_iso2022",
    "_codecs_jp",
    "_codecs_kr",
    "_codecs_tw",
    "_collections_abc",
    "_compat_pickle",
    "_compression",
    "_crypt",
    "_csv",
    "_ctypes",
    "_ctypes_test",
    "_curses",
    "_curses_panel",
    "_dbm",
    "_decimal",
    "_dummy_thread",
    "_elementtree",
    "_gdbm",
    "_hashlib",
    "_hotshot",
    "_json",
    "_lsprof",
    "_LWPCookieJar",
    "_lzma",
    "_markupbase",
    "_MozillaCookieJar",
    "_multibytecodec",
    "_multiprocessing",
    "_opcode",
    "_osx_support",
    "_pydecimal",
    "_pyio",
    "_sitebuiltins",
    "_sqlite3",
    "_ssl",
    "_strptime",
    "_sysconfigdata",
    "_sysconfigdata_m",
    "_sysconfigdata_nd",
    "_testbuffer",
    "_testcapi",
    "_testimportmultiple",
    "_testmultiphase",
    "_threading_local",
    "_tkinter",
    "_weakrefset",
    "abc",
    "aifc",
    "antigravity",
    "anydbm",
    "argparse",
    "ast",
    "asynchat",
    "asyncio",
    "asyncore",
    "atexit",
    "audiodev",
    "audioop",
    "base64",
    "BaseHTTPServer",
    "Bastion",
    "bdb",
    "binhex",
    "bisect",
    "bsddb",
    "bz2",
    "calendar",
    "Canvas",
    "CDROM",
    "cgi",
    "CGIHTTPServer",
    "cgitb",
    "chunk",
    "cmath",
    "cmd",
    "code",
    "codecs",
    "codeop",
    "collections",
    "colorsys",
    "commands",
    "compileall",
    "compiler",
    "concurrent",
    "ConfigParser",
    "configparser",
    "contextlib",
    "Cookie",
    "cookielib",
    "copy",
    "copy_reg",
    "copyreg",
    "cProfile",
    "crypt",
    "csv",
    "ctypes",
    "curses",
    "datetime",
    "dbhash",
    "dbm",
    "decimal",
    "Dialog",
    "difflib",
    "dircache",
    "dis",
    "distutils",
    "DLFCN",
    "doctest",
    "DocXMLRPCServer",
    "dumbdbm",
    "dummy_thread",
    "dummy_threading",
    "easy_install",
    "email",
    "encodings",
    "ensurepip",
    "enum",
    "filecmp",
    "FileDialog",
    "fileinput",
    "FixTk",
    "fnmatch",
    "formatter",
    "fpectl",
    "fpformat",
    "fractions",
    "ftplib",
    "functools",
    "future_builtins",
    "genericpath",
    "getopt",
    "getpass",
    "gettext",
    "glob",
    "gzip",
    "hashlib",
    "heapq",
    "hmac",
    "hotshot",
    "html",
    "htmlentitydefs",
    "htmllib",
    "HTMLParser",
    "http",
    "httplib",
    "idlelib",
    "ihooks",
    "imaplib",
    "imghdr",
    "imp",
    "importlib",
    "imputil",
    "IN",
    "inspect",
    "io",
    "ipaddress",
    "json",
    "keyword",
    "lib2to3",
    "linecache",
    "linuxaudiodev",
    "locale",
    "logging",
    "lzma",
    "macpath",
    "macurl2path",
    "mailbox",
    "mailcap",
    "markupbase",
    "md5",
    "mhlib",
    "mimetools",
    "data",
    "MimeWriter",
    "mimify",
    "mmap",
    "modulefinder",
    "multifile",
    "multiprocessing",
    "mutex",
    "netrc",
    "new",
    "nis",
    "nntplib",
    "ntpath",
    "nturl2path",
    "numbers",
    "opcode",
    "operator",
    "optparse",
    "os",
    "os2emxpath",
    "ossaudiodev",
    "parser",
    "pathlib",
    "pdb",
    "pickle",
    "pickletools",
    "pip",
    "pipes",
    "pkg_resources",
    "pkgutil",
    "platform",
    "plistlib",
    "popen2",
    "poplib",
    "posixfile",
    "posixpath",
    "pprint",
    "profile",
    "pstats",
    "pty",
    "py_compile",
    "pyclbr",
    "pydoc",
    "pydoc_data",
    "pyexpat",
    "Queue",
    "queue",
    "quopri",
    "random",
    "re",
    "readline",
    "repr",
    "reprlib",
    "resource",
    "rexec",
    "rfc822",
    "rlcompleter",
    "robotparser",
    "runpy",
    "sched",
    "ScrolledText",
    "selectors",
    "sets",
    "setuptools",
    "sgmllib",
    "sha",
    "shelve",
    "shlex",
    "shutil",
    "signal",
    "SimpleDialog",
    "SimpleHTTPServer",
    "SimpleXMLRPCServer",
    "site",
    "sitecustomize",
    "smtpd",
    "smtplib",
    "sndhdr",
    "socket",
    "SocketServer",
    "socketserver",
    "sqlite3",
    "sre",
    "sre_compile",
    "sre_constants",
    "sre_parse",
    "ssl",
    "stat",
    "statistics",
    "statvfs",
    "string",
    "StringIO",
    "stringold",
    "stringprep",
    "struct",
    "subprocess",
    "sunau",
    "sunaudio",
    "symbol",
    "symtable",
    "sysconfig",
    "tabnanny",
    "tarfile",
    "telnetlib",
    "tempfile",
    "termios",
    "test",
    "textwrap",
    "this",
    "threading",
    "timeit",
    "Tix",
    "tkColorChooser",
    "tkCommonDialog",
    "Tkconstants",
    "Tkdnd",
    "tkFileDialog",
    "tkFont",
    "tkinter",
    "Tkinter",
    "tkMessageBox",
    "tkSimpleDialog",
    "toaiff",
    "token",
    "tokenize",
    "trace",
    "traceback",
    "tracemalloc",
    "ttk",
    "tty",
    "turtle",
    "types",
    "TYPES",
    "typing",
    "unittest",
    "urllib",
    "urllib2",
    "urlparse",
    "user",
    "UserDict",
    "UserList",
    "UserString",
    "uu",
    "uuid",
    "venv",
    "warnings",
    "wave",
    "weakref",
    "webbrowser",
    "wheel",
    "whichdb",
    "wsgiref",
    "xdrlib",
    "xml",
    "xmllib",
    "xmlrpc",
    "xmlrpclib",
    "xxlimited",
    "zipapp",
    "zipfile",
}

module = type(sys)
PEP440_VERSION_PATTERN = re.compile(
    r"""
^
(\d+!)?              # epoch
(\d[.\d]*(?<= \d))   # release
((?:[abc]|rc)\d+)?   # pre-release
(?:(\.post\d+))?     # post-release
(?:(\.dev\d+))?      # development release
$
""",
    flags=re.VERBOSE,
)


def join_paths(*parts):
    """Join different parts together to a valid dotted path."""
    return ".".join(str(p).strip(".") for p in parts if p)


def iter_prefixes(path):
    """
    Iterate through all (non-empty) prefixes of a dotted path.

    Example
    -------
    >>> list(iter_prefixes('foo.bar.baz'))
    ['foo', 'foo.bar', 'foo.bar.baz']
    """
    split_path = path.split(".")
    for i in range(1, len(split_path) + 1):
        yield join_paths(*split_path[:i])


def optional_import(*package_names):
    try:
        packages = [importlib.import_module(pn) for pn in package_names]
        return True, packages[0]
    except ImportError:
        return False, None


def get_py_file_if_possible(pyc_name):
    """Try to retrieve a X.py file for a given X.py[c] file."""
    if pyc_name.endswith((".py", ".so", ".pyd")):
        return pyc_name
    assert pyc_name.endswith(".pyc")
    non_compiled_file = pyc_name[:-1]
    if os.path.exists(non_compiled_file):
        return non_compiled_file
    return pyc_name


def get_digest(filename):
    """Compute the MD5 hash for a given file."""
    h = hashlib.md5()
    with open(filename, "rb") as f:
        data = f.read(1 * MB)
        while data:
            h.update(data)
            data = f.read(1 * MB)
        return h.hexdigest()


def get_commit_if_possible(filename, save_git_info):
    """Try to retrieve VCS information for a given file.

    Currently only supports git using the gitpython package.

    Parameters
    ----------
    filename : str

    Returns
    -------
        path: str
            The base path of the repository
        commit: str
            The commit hash
        is_dirty: bool
            True if there are uncommitted changes in the repository
    """
    if save_git_info is False:
        return None, None, None

    directory = os.path.dirname(filename)
    try:
        repo = Repo(directory, search_parent_directories=True)
    except InvalidGitRepositoryError:
        return None, None, None
    try:
        path = repo.remote().url
    except ValueError:
        path = "git:/" + repo.working_dir
    is_dirty = repo.is_dirty()
    commit = repo.head.commit.hexsha
    return path, commit, is_dirty


@functools.total_ordering
class Source:
    def __init__(self, filename, digest, repo, commit, isdirty):
        self.filename = filename
        self.digest = digest
        self.repo = repo
        self.commit = commit
        self.is_dirty = isdirty

    @staticmethod
    def create(filename, save_git_info=True):
        if not filename or not os.path.exists(filename):
            raise ValueError('invalid filename or file not found "{}"'.format(filename))

        main_file = get_py_file_if_possible(os.path.abspath(filename))
        repo, commit, is_dirty = get_commit_if_possible(main_file, save_git_info)
        return Source(main_file, get_digest(main_file), repo, commit, is_dirty)

    def to_json(self, base_dir=None):
        source = {}
        if base_dir:
            return os.path.relpath(self.filename, base_dir), self.digest
        else:
            source.setdefault(LcConst.SOURCE_FILE_NAME, self.filename)
            source.setdefault(LcConst.SOURCE_FILE_DIGEST, self.digest)
            source.setdefault(LcConst.SOURCE_FILE_GIT_REPO, self.repo)
            source.setdefault(LcConst.SOURCE_FILE_GIT_COMMIT_ID, self.commit)
            source.setdefault(LcConst.SOURCE_FILE_IS_DIRTY, self.is_dirty)
            source.setdefault(LcConst.SOURCE_FILE_PATH, os.path.dirname(self.filename))
            return source
            # return self.filename, self.digest

    def __hash__(self):
        return hash(self.filename)

    def __eq__(self, other):
        if isinstance(other, Source):
            return self.filename == other.filename
        elif isinstance(other, str):
            return self.filename == other
        else:
            return False

    def __le__(self, other):
        return self.filename.__le__(other.filename)

    def __repr__(self):
        return "Source: {}".format(self.filename)


@functools.total_ordering
class PackageDependency:
    modname_to_dist = {}

    def __init__(self, name, version):
        self.name = name
        self.version = version
        self.json_data = None

    def fill_missing_version(self):
        if self.version is not None:
            return
        dist = pkg_resources.working_set.by_key.get(self.name)
        self.version = dist.version if dist else None

    def to_json(self):
        str_value="\"{}\":\"{}\"".format(self.name, self.version or "<unknown>")
        self.json_data = "{"+ str_value + "}"
        return self

    def to_dict(self):
        return json.loads(self.json_data)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, PackageDependency):
            return self.name == other.name
        else:
            return False

    def __le__(self, other):
        return self.name.__le__(other.name)

    def __repr__(self):
        return "<PackageDependency: {}={}>".format(self.name, self.version)

    @classmethod
    def create(cls, mod):
        if not cls.modname_to_dist:
            # some packagenames don't match the module names (e.g. PyYAML)
            # so we set up a dict to map from module name to package name
            for dist in pkg_resources.working_set:
                try:
                    toplevel_names = dist._get_metadata("top_level.txt")
                    for tln in toplevel_names:
                        cls.modname_to_dist[tln] = dist.project_name, dist.version
                except Exception:
                    pass

        name, version = cls.modname_to_dist.get(mod.__name__, (mod.__name__, None))

        return PackageDependency(name, version)


def convert_path_to_module_parts(path):
    """Convert path to a python file into list of module names."""
    module_parts = list(path.parts)
    if module_parts[-1] in ["__init__.py", "__init__.pyc"]:
        # remove trailing __init__.py
        module_parts = module_parts[:-1]
    else:
        # remove file extension
        module_parts[-1], _ = os.path.splitext(module_parts[-1])
    return module_parts


def is_local_source(filename, modname, experiment_path):
    """Check if a module comes from the given experiment path.

    Check if a module, given by name and filename, is from (a subdirectory of )
    the given experiment path.
    This is used to determine if the module is a local source file, or rather
    a package dependency.

    Parameters
    ----------
    filename: str
        The absolute filename of the module in question.
        (Usually module.__file__)
    modname: str
        The full name of the module including parent namespaces.
    experiment_path: str
        The base path of the experiment.

    Returns
    -------
    bool:
        True if the module was imported locally from (a subdir of) the
        experiment_path, and False otherwise.
    """
    filename = Path(os.path.abspath(os.path.realpath(filename)))
    experiment_path = Path(os.path.abspath(os.path.realpath(experiment_path)))
    if experiment_path not in filename.parents:
        return False
    rel_path = filename.relative_to(experiment_path)
    path_parts = convert_path_to_module_parts(rel_path)

    mod_parts = modname.split(".")
    if path_parts == mod_parts:
        return True
    if len(path_parts) > len(mod_parts):
        return False
    abs_path_parts = convert_path_to_module_parts(filename)
    return all([p == m for p, m in zip(reversed(abs_path_parts), reversed(mod_parts))])


def get_main_file(globs, save_git_info):
    filename = globs.get("__file__")

    if filename is None:
        experiment_path = os.path.abspath(os.path.curdir)
        main = None
    else:
        main = Source.create(globs.get("__file__"), save_git_info)
        experiment_path = os.path.dirname(main.filename)
        main.to_json()

    return experiment_path, main


def iterate_imported_modules(globs):
    checked_modules = set(MODULE_BLACKLIST)
    for glob in globs.values():
        if isinstance(glob, module):
            mod_path = glob.__name__
        elif hasattr(glob, "__module__"):
            mod_path = glob.__module__
        else:
            continue  # pragma: no cover

        if not mod_path:
            continue

        for modname in iter_prefixes(mod_path):
            if modname in checked_modules:
                continue
            checked_modules.add(modname)
            mod = sys.modules.get(modname)
            if mod is not None:
                yield modname, mod


def iterate_all_python_files(base_path):
    # TODO support ignored directories/files
    for dirname, subdirlist, filelist in os.walk(base_path):
        if "__pycache__" in dirname:
            continue
        for filename in filelist:
            if filename.endswith(".py"):
                yield os.path.join(base_path, dirname, filename)


def iterate_sys_modules():
    items = list(sys.modules.items())
    for modname, mod in items:
        if modname not in MODULE_BLACKLIST and mod is not None:
            yield modname, mod


def get_sources_from_modules(module_iterator, base_path, save_git_info):
    sources = set()
    for modname, mod in module_iterator:
        # hasattr doesn't work with python extensions
        if not getattr(mod, "__file__", None):
            continue

        filename = os.path.abspath(mod.__file__)
        if filename not in sources and is_local_source(filename, modname, base_path):
            s = Source.create(filename, save_git_info)
            sources.add(s)

    return sources


def get_dependencies_from_modules(module_iterator, base_path):
    dependencies = {}
    for modname, mod in module_iterator:
        # hasattr doesn't work with python extensions
        if getattr(mod, "__file__", None) and is_local_source(
            os.path.abspath(mod.__file__), modname, base_path
        ):
            continue
        if modname.startswith("_") or "." in modname:
            continue

        try:
            pdep = PackageDependency.create(mod)
            if pdep.version is not None:
                pdep=pdep.to_json().to_dict()
                dependencies.update(pdep)
        except AttributeError:
            pass
    return dependencies


def get_sources_from_sys_modules(globs, base_path, save_git_info):
    return get_sources_from_modules(iterate_sys_modules(), base_path, save_git_info)


def get_sources_from_imported_modules(globs, base_path, save_git_info):
    return get_sources_from_modules(
        iterate_imported_modules(globs), base_path, save_git_info
    )


def get_sources_from_local_dir(globs, base_path, save_git_info):
    return {
        Source.create(filename, save_git_info)
        for filename in iterate_all_python_files(base_path)
    }


def get_dependencies_from_sys_modules(globs, base_path):
    return get_dependencies_from_modules(iterate_sys_modules(), base_path)


def get_dependencies_from_imported_modules(globs, base_path):
    return get_dependencies_from_modules(iterate_imported_modules(globs), base_path)


def get_dependencies_from_pkg(globs, base_path):
    dependencies = set()
    for dist in pkg_resources.working_set:
        if dist.version == "0.0.0":
            continue  # ugly hack to deal with pkg-resource version bug
        dependencies.add(PackageDependency(dist.project_name, dist.version))
    return dependencies


source_discovery_strategies = {
    "none": lambda globs, path: set(),
    "imported": get_sources_from_imported_modules,
    "sys": get_sources_from_sys_modules,
    "dir": get_sources_from_local_dir,
}

dependency_discovery_strategies = {
    "none": lambda globs, path: set(),
    "imported": get_dependencies_from_imported_modules,
    "sys": get_dependencies_from_sys_modules,
    "pkg": get_dependencies_from_pkg,
}


def gather_sources_and_dependencies(globs, save_git_info, base_dir=None):
    """Scan the given globals for modules and return them as dependencies."""
    experiment_path, main = get_main_file(globs, save_git_info)

    base_dir = base_dir or experiment_path

    # brian_tag call get_sources_from_imported_modules() == gather_sources
    gather_sources = source_discovery_strategies[SETTINGS["DISCOVER_SOURCES"]]
    sources = gather_sources(globs, base_dir, save_git_info)
    if main is not None:
        sources.add(main)

    # brian_tag 'imported' is used.
    # get_dependencies_from_imported_modules() == gather_dependencies
    gather_dependencies = dependency_discovery_strategies[
        SETTINGS["DISCOVER_DEPENDENCIES"]
    ]
    dependencies = gather_dependencies(globs, base_dir)

    has_numpy, np = optional_import("numpy")
    if has_numpy:
        # Add numpy as a dependency because it might be used for randomness
        np_value = PackageDependency.create(np)
        np_value = np_value.to_json().to_dict()
        dependencies.update(np_value)

    main_json = {}
    if main is not None:
        main_json = main.to_json()
    return main_json, sources, dependencies




import os
import inspect
import json
import datetime
from collections import OrderedDict
import warnings
import sys
import logging
import boto3
from boto3.session import Session
import asyncio
from slack import WebClient
from slack.errors import SlackApiError
from pathlib import Path
from joblib import dump
from Accuinsight.modeler.core import func, path, get
from Accuinsight.modeler.core.func import get_time
from Accuinsight.modeler.core.get_for_visual import roc_pr_curve, get_visual_info_regressor, get_true_y, get_visual_info_regressor
from Accuinsight.modeler.utils.runs_utils import get_aws_info, ProgressPercentage
from Accuinsight.modeler.core.sklearnModelType import REGRESSION, CLASSIFICATION
from Accuinsight.modeler.core.LcConst.LcConst import ALL_MODEL_PARAMS, SELECTED_PARAMS, SELECTED_METRICS, VALUE_ERROR, \
    LOGGING_TIME, LOGGING_RUN_ID, FITTED_MODEL, RUN_PREFIX_PATH, RUN_MODEL_JSON_PATH, RUN_MODEL_VISUAL_JSON_PATH, \
    RUN_MODEL_HOME_PATH, ENV_JUPYTER_HOME_DIR
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import set_current_runs, clear_runs, print_run_info, \
    set_git_meta, set_python_dependencies, set_run_name, set_model_json_path, set_visual_json_path, set_model_file_path, set_prefix_path, set_best_model_joblib_path, _set_result_path, set_shap_json_path 
from Accuinsight.modeler.utils.dependency.dependencies import gather_sources_and_dependencies
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.utils.os_getenv import is_in_ipython, get_current_notebook
from Accuinsight.modeler.clients.modeler_api import LifecycleRestApi
from Accuinsight.modeler.core.feature_contribution import shap_value 


logging.basicConfig(level=logging.INFO,
                    format='%(message)s')

warnings.filterwarnings("ignore")


class accuinsight(object):
    def __init__(self):
        self.BucketInfo = None
        self.endpoint = None

    def get_file(self, aws_json_file_name):
        global save_path, target_name, bucket_info
        
        # meta file 저장하는 폴더 만들면, 이 경로는 수정돼야 함.
        aws_info_json_path = os.path.join(ENV_JUPYTER_HOME_DIR, aws_json_file_name)
    
        with open(aws_info_json_path) as jsonFile:
            self.BucketInfo = json.load(jsonFile)
            
        BUCKET_TYPE = self.BucketInfo['bucketType']
        BUCKET_NAME = self.BucketInfo['bucketName']
        FILE_PATH = self.BucketInfo['filePath']
        FILE_NAME = self.BucketInfo['fileName']
        #FILE_TYPE = self.BucketInfo['fileType']
        #FILE_DELIM = self.BucketInfo['fileDelim']
        ACCESS_KEY = self.BucketInfo['myAccessKey']
        SECRET_KEY = self.BucketInfo['mySecretKey']
        REGION = self.BucketInfo['region']
        URL = self.BucketInfo['endpoint']
        
        ## define the target name 
        target_name = self.BucketInfo['target']

        ## path for saving data 
        save_dir = os.path.join(ENV_JUPYTER_HOME_DIR, 'data_from_AWS')  

        if os.path.exists(save_dir) == False:
            os.mkdir(save_dir)
        else:
            pass
  
        self.save_file_name = FILE_NAME.split('.')[0] + '_' + str(datetime.datetime.now().date()).replace('-', '') + '.' + FILE_NAME.split('.')[1]
        save_path = os.path.join(save_dir, self.save_file_name)

        # endpoint
        pre_url = 'https://' + BUCKET_NAME + '.' + URL
        self.endpoint = os.path.join(pre_url, FILE_PATH)

        client = boto3.client(BUCKET_TYPE,
                              aws_access_key_id=ACCESS_KEY,
                              aws_secret_access_key=SECRET_KEY,
                              region_name=REGION)

        transfer = boto3.s3.transfer.S3Transfer(client)

        progress = ProgressPercentage(client, BUCKET_NAME, FILE_PATH)

        sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME, '\n'))
        transfer.download_file(BUCKET_NAME, FILE_PATH, save_path, callback=progress)
        logging.info(save_path)
        
        bucket_info = self.BucketInfo
    
    def set_slack(self, token = None, channel_id = None):
        self.token = token
        self.channel_id = channel_id

    def send_message(self, message = None):
        if message is not None:
            try:
                message = message
                response = WebClient(token=self.token).chat_postMessage(channel=self.channel_id, text=message)
            except SlackApiError as e:
                assert e.response["error"]
        else:
            raise ValueError('message를 입력해주세요.')


    class add_experiment(object):
        def __init__(self, model_name, *args, model_monitor=False):
            if model_monitor == True:
                self.shap_on = True
                logging.info('Using add_experiment(model_monitor=True)')
            else:
                self.shap_on = False
                logging.info('Using add_experiment(model_monitor=False)')
            
            self.model_name = model_name
        
            is_notebook = is_in_ipython()
            if is_notebook == True:
                self.var_model_file_path = get_current_notebook()
                _caller_globals = inspect.stack()[1][0].f_globals
                (
                    self.mainfile,
                    self.sources,
                    self.dependencies
                ) = gather_sources_and_dependencies(
                    globs=_caller_globals,
                    save_git_info=False
                )
            else:
                _caller_globals = inspect.stack()[1][0].f_globals
                (
                    self.mainfile,
                    self.sources,
                    self.dependencies
                ) = gather_sources_and_dependencies(
                    globs=_caller_globals,
                    save_git_info=True
                )
                self.var_model_file_path = self.mainfile['filename']

            self.fitted_model = get.model_type(self.model_name)          # fitted model type
            self.json_path = OrderedDict()                               # path
            self.selected_params = []                                    # log params
            self.selected_metrics = OrderedDict()                        # log metrics
            self.summary_info = OrderedDict()                            # final results
            self.error_log = []                                          # error log
            self.vis_info = None                                         # visualization info - classifier
            self.dict_path = path.get_file_path(self.model_name)

            set_current_runs(get.model_type(self.model_name))
            _set_result_path(self.dict_path)

            # visualization function
            if len(args) == 2:
                self.xval = args[0]
                self.yval = args[1]

            elif len(args) == 4:
                self.xtrain = args[0]
                self.ytrain = args[1]
                self.xval = args[2]
                self.yval = args[3]
    
            else:
                raise ValueError('Check the arguments of function - add_experiment(model_name, X_val, y_val) or add_experiment(model_name, X_train, y_train, X_val, y_val)')

            # classifier
            if any(i in self.fitted_model for i in CLASSIFICATION):
                self.vis_info = roc_pr_curve(self.model_name, self.xval, self.yval)

            # regressor
            elif any(i in self.fitted_model for i in REGRESSION):
                self.ypred = get_visual_info_regressor(self.model_name, self.xval)
                
            # sklearn/xgboost/lightgbm
            get_from_model = get.from_model(self.model_name)
            self.all_model_params = get_from_model.all_params()
            self.model_param_keys = get_from_model.param_keys()

            set_model_file_path(self.var_model_file_path)

            if hasattr(self, 'mainfile'):
                set_git_meta(fileinfo=self.mainfile)
            if hasattr(self, 'dependencies'):
                set_python_dependencies(py_depenpency=self.dependencies)

        def __enter__(self):
            self.start_time = get_time.now()
            self.summary_info[LOGGING_TIME] = get_time.logging_time()
            self.summary_info[LOGGING_RUN_ID] = func.get_run_id()
            self.summary_info[FITTED_MODEL] = self.fitted_model

            set_prefix_path(self.dict_path[LcConst.RUN_PREFIX_PATH])
            set_run_name(self.fitted_model, self.summary_info[LOGGING_RUN_ID])

            return(self)

        def __exit__(self, a, b, c):
            self.summary_info[ALL_MODEL_PARAMS] = self.all_model_params
            self.summary_info[SELECTED_PARAMS] = self.selected_params
            self.summary_info[SELECTED_METRICS] = self.selected_metrics
            self.summary_info[VALUE_ERROR] = self.error_log

            # model_monitor = True
            self.run_id = self.summary_info[LOGGING_RUN_ID]
            
            if self.shap_on == True:
                try:
                    self.feature_name = get.feature_name(save_path, bucket_info, target_name)
                except:
                    self.feature_name = None
                self.run_id = self.fitted_model + '-' + self.run_id
                
                self.shap_value = shap_value(self.model_name, self.xtrain, self.feature_name)
                
                # path for shap.json
                shap_json_full_path = self.dict_path['shap_json_full']
                set_shap_json_path(self.dict_path['shap_json'])
                
                with open(shap_json_full_path, 'w', encoding = 'utf-8') as save_file:
                    json.dump(self.shap_value, save_file, indent='\t')
                
            else:   
                pass
                
            # visualization
            if any(i in self.fitted_model for i in CLASSIFICATION):

                # path for visual.json
                visual_json_full_path = self.dict_path['visual_json_full']
                set_visual_json_path(self.dict_path['visual_json'])

                with open(visual_json_full_path, 'w', encoding='utf-8') as save_file1:
                    json.dump(self.vis_info, save_file1, indent="\t")

            elif any(i in self.fitted_model for i in REGRESSION):
                temp_yval = get_true_y(self.yval)
                if len(temp_yval) <= 5000:
                    self.summary_info['True_y'] = temp_yval
                    self.summary_info['Predicted_y'] = get_visual_info_regressor(self.model_name, self.xval)
                else:
                    self.summary_info['True_y'] = None
                    self.summary_info['Predicted_y'] = None

            self.summary_info['ValueError'] = self.error_log

            if not self.summary_info['ValueError']:
                # path for model_info.json
                model_json_full_path = self.dict_path['model_json_full']
                set_model_json_path(self.dict_path['model_json'])

                with open(model_json_full_path, 'w', encoding='utf-8') as save_file2:
                    json.dump(self.summary_info, save_file2, indent="\t")
            else:
                pass

            # model save
            save_model_path = self.dict_path['save_model_joblib'] + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID] +'.joblib'
            path_for_setting_model_joblib = self.dict_path['save_model_dir'] + '/' + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID] +'.joblib'

            set_best_model_joblib_path(path_for_setting_model_joblib)

            dump(self.model_name, save_model_path)

            start_time = int(self.start_time.timestamp()*1000)
            end_time = int(get_time.now().timestamp()*1000)
            delta_ts = end_time - start_time

            clear_runs(start_time, end_time, delta_ts)
            modeler_rest = LifecycleRestApi(LcConst.BACK_END_API_URL,
                                            LcConst.BACK_END_API_PORT,
                                            LcConst.BACK_END_API_URI)
            modeler_rest.lc_create_run()


        def log_params(self, param = None):
            # sklearn/xgboost/lightgbm
            if param:
                
                if param in self.model_param_keys:
                    return(self.selected_params.append(param))

                else:
                    self.error_log.append(True)
                    raise ValueError('"' + param + '"' + ' does not exist in the model.')

        def log_metrics(self, metric_name, defined_metric):
            self.selected_metrics[metric_name] = defined_metric

        def log_tag(self, description):
            self.summary_info['tag'] = description



import argparse
import subprocess
from Accuinsight.modeler.clients.modeler_api import WorkspaceRunRestApi
from Accuinsight.modeler.entities.workspace_run_log import WorkspaceRunLog
from Accuinsight.modeler.core.LcConst import LcConst


class WorkspaceRun:
    """
        Object for running code and sending the result to backend.
    """
    def __init__(self):
        self.workspace_run_log = WorkspaceRunLog()
        self.workspace_run_api = WorkspaceRunRestApi(LcConst.BACK_END_API_URL,
                                                     LcConst.BACK_END_API_PORT,
                                                     LcConst.BACK_END_API_URI)
        self.code_path = None

    def exec_code(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--workspaceRunId', default=None)
        parser.add_argument('--codePath', default=None)
        parser.add_argument('--stopFlag', default=False)
        parser.add_argument('--stopTimeout', default=60)

        args, unknown = parser.parse_known_args()
        args_dict = vars(args)

        self.code_path = args_dict['codePath']
        self.workspace_run_log.workspace_run_id = args_dict['workspaceRunId']
        self.workspace_run_log.stop_flag = args_dict['stopFlag']
        self.workspace_run_log.stop_timeout = args_dict['stopTimeout']

        if not self.code_path:
            raise Exception("codePath cannot be none")

        try:
            subprocess.run("python3 -u %s > "
                           "/tmp/output_%s.log 2>&1"
                           % (self.code_path, self.workspace_run_log.workspace_run_id), shell=True, encoding='UTF-8')\
                .check_returncode()
            self.workspace_run_log.is_success = True
        except subprocess.CalledProcessError:
            self.workspace_run_log.is_success = False
        finally:
            self.workspace_run_api.call_rest_api(self.workspace_run_log.get_result_param())


if __name__ == "__main__":
    workspace_run = WorkspaceRun()
    workspace_run.exec_code()



# progressbar in boto3
import threading
import sys
import math

class ProgressPercentage(object):
    def __init__(self, client, bucket, filename):
        self._filename = filename
        self._size = client.head_object(Bucket=bucket, Key=filename).get('ContentLength')
        self._seen_so_far = 0
        self._lock = threading.Lock()
    
    def __call__(self, bytes_amount):
        def convertSize(size):
            if (size == 0):
                return '0B'
            size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
            i = int(math.floor(math.log(size,1024)))
            p = math.pow(1024,i)
            s = round(size/p,2)
            return '%.2f %s' % (s,size_name[i])

        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            done = int(50 * self._seen_so_far / self._size)
            sys.stdout.write(
                    "\r[%s%s] %s/%s  (%.2f%%)        " % (
                    '=' * done, ' ' * (50-done) , convertSize(self._seen_so_far), convertSize(self._size),
                    percentage))

            sys.stdout.flush()

# get awsInfo from datacatalog
import os
import json
from pathlib import Path
from collections import OrderedDict
import requests


def get_aws_info(sub_path):
    host = 'https://accuinsight-dev.cloudz.co.kr/datacatalogapi/'
    
    headers = {'Content-Type': 'application/json', 'userid': 'bigdata_poc'} # 2차 고도화 예정
    
    url = host + sub_path
    
    response = requests.request("GET", url, headers=headers, verify=False)
    awsInfo = json.loads(response.content)
    
    connectInfo = OrderedDict()
    
    connectInfo['my_access_key'] = awsInfo['connection']['credential']['accessKey']
    connectInfo['my_secret_key'] = awsInfo['connection']['credential']['secretKey']
    connectInfo['endpoint'] = awsInfo['connection']['url']
    
    splitEndpoint = awsInfo['location'].split('/')[2:]

    connectInfo['region'] = awsInfo['connection']['url'].split('.')[1]
    connectInfo['bucket_type'] = awsInfo['connection']['type']
    connectInfo['bucket_name'] = splitEndpoint[0]
    connectInfo['file_path'] = os.path.join(*splitEndpoint[1:])
    connectInfo['file_name'] = splitEndpoint[-1]
    connectInfo['file_type'] = awsInfo['dataSourceType']
    connectInfo['file_delim'] = awsInfo['dataSourceDelimiter']
    
    return(connectInfo)
